#!/usr/bin/tclsh
#
# v_sim.tcl, Secure Incoming Mover
#
set revision "Apr/8/2002"

#
#
# :: Description ::
#     This script is ideal for sites that have large encrypted
#     filesystems but want races to be performed on an unencrypted area
#     (such as an /incoming/ dir).  Once the dir has not been uploaded
#     new files after N minutes, it is copied to its respective location
#     in an encrypted filesystem and finally wiped from its original,
#     unencrypted location.  English?  Say you collect only apps.  You
#     have a big raid (/dev/md0) mounted as /glftpd/site/apps/ which
#     uses an encrypted filesystem.  You also have a 15G (/dev/hdb1)
#     mounted as /glftpd/site/incoming/apps/ which is not encrypted.
#     Users are only allowed to upload to the /incoming/ branch (because
#     you configured glftpd.conf that way) and once the dir hasn't been
#     touched for N minutes (say half an hour) it is securely moved to
#     /apps/ from /incoming/apps/.  I somewhat assume you've read my
#     encryption paper "howto: hard drive encryption for your linux ftp
#     site".  Grab it from my homepage and this will all make more
#     sense.
#
# :: Requirements ::
#     1) unix utils: uptime, du
#     2) bcwipe from http://www.jetico.com (it's free)
#     These utils should be accessible by your PATH environment
#     variable (/glftpd/bin/ is a good choice).
#
# :: Installation ::
#     Configure below and test it manually, then cronjob it:
#         */15 * * * *        /glftpd/bin/v_sim.tcl 2>&1 >/dev/null
#
# :: Exit codes ::
#     0 - successful run
#     1 - some arbitrary tcl error
#     2 - failed to commence because system load is too high
#     3 - movetable is not configured right (not a real dir)
#
# :: Todo ::
# 1. add a proc to represent all times in human readable HMS
# 2. extra function for removeDupeDest to set to '2'.  when
#    dupe exists, still copy it but rename the dir as you
#    copy with some extra chars at end (random?).  then nukers
#    can assess the situation later.  if user can't delete
#    anything inside the dest area... then this forces a nuke.
#    it's good because the user will remember to check the dir
#    "from then on", plus, when there is a race... you know
#    what is new and what is old.  you shouldn't be duping
#    an entire release anyhow, unless it's a mass backfill.
# 3. another method of defining when a release is "old enough"
#    to be moved.  in addition to mtime...use dir/file match
#    like [_completed_]...etc. from zipscripts.
# 4. lock down source dir while it's being bcwiped or worked on in
#    general?
#
#######################################################################


### Configuration ###
#
# If the current system load (from `uptime`) is greater in any of the
# three values (for 1, 5, 15 minutes), then bail out and don't move
# anything.  If you don't want a requirement, then it's safe to put
# a large number such as 99.00.  Do `man uptime` for more information.
# The first number below is for the 1 minute load, then 5 and 10.
set maxSystemLoad {
    0.20
    99.00
    99.00
}

# To enable set to 1, to disable set to 0.  If a dir is going to be
# moved and its destination already exists, this will delete
# the destination and then proceed with moving the source it has.
# Note that "moving" entails two steps: copy, then bcwipe the original.
# This may be useful if for some reason your drive is filling up
# rapidly and the free space checker sees enough space for the destination,
# then the release is copied...but the destination partition fills
# before the release is entirely copied (and thereafter bcwiped).
# So you have a partial release in the destination dir.  This
# ensures the destination contains the "full" release, in theory. ;)
# If disabled and a duplicate destination exists, then the source
# will be skipped and a warning is output to stdout.  Now if someone
# upload a DUPE, then you obviously don't want this set!  Else you'd
# delete your old "good" release by a new dupe...so this is disabled
# by default.  The first case mentioned won't happen as often as the
# second one, possibly.  How's that for confidence? :P
set removeDupeDest 0

# If a dir's modification time (mtime) was at least this
# many minutes ago, then move it.
set releaseMoveTime 30

# If destination does not have enough free space for the
# dir which we want to move, then run this script to free
# up some space, then attempt to move the release again.
set freeSpaceScript "/glftpd/bin/v_free.tcl"

# If the dir we want to move matches any string in this list,
# do nothing.  Each string must be on its own line in double-quotes.
# You may use '*' wildcard.
set skipList {
    "*lost+found"
    "*nuked*"
}

# If this file/dir exists in the releases dir, it will not be moved.
# So if dir is /glftpd/site/incoming/utils/My_Release,
# mkdir /glftpd/site/incoming/utils/My_Release/<skipTag> or
# touch /glftpd/site/incoming/utils/My_Release/<skipTag>.
set skipTag "keep"

# <srcDir:destDir>
# Look for dirs to move from srcDir/* to destDir.  Observe the
# colon separator.
set moveTable {
    "/glftpd/site/incoming/tv:/glftpd/site/tv"
    "/glftpd/site/incoming/games:/glftpd/site/games"
    "/glftpd/site/incoming/console/ps2_dvd:/glftpd/site/console/ps2_dvd"
    "/glftpd/site/incoming/console/xbox:/glftpd/site/console/xbox"
    "/glftpd/site/incoming/movies/vcd:/glftpd/site/movies/vcd"
    "/glftpd/site/incoming/movies/svcd:/glftpd/site/movies/svcd"
    "/glftpd/site/incoming/movies/svcd-classic:/glftpd/site/movies/svcd-classic"
}

#######################################################################




# Define exit codes.
set exit(SUCCESS) 0
set exit(HIGH_LOAD) 2
set exit(BAD_MOVETABLE) 3

# Returns 1 if the incoming dir "is old enough" as defined
# by he variable releaseMoveTime, else, returns 0.  Time
# is measured in the dir's modification time.
proc isOldEnough {dir} {
    global releaseMoveTime
    # Multi-cd releases must be subdir'ed with "CD<#>".
    set subDirs [glob -nocomplain -type d -directory $dir "cd*" "Cd*" "CD*" "cD*"]

    # If it is a multi-cd release, base the age on the CD# dirs only.
    if {[llength $subDirs] > 0} {
        set dirsToCheck $subDirs
    } else {
        # If not, just go by the release dir.
        set dirsToCheck $dir
    }
    # If the release is multi-cd, then all cds must "pass" the age test,
    # else there is some dir that is "new" and that makes the entire release
    # disqualified to be moved.  The logic allows non multi-cd releases to
    # be processed too. :-)
    set allDirsPass 0
    foreach myDir $dirsToCheck {
        # Grab the age of the dir, in minutes, as of now.
        set ageOfDir [expr ([clock seconds] - [file mtime $myDir]) / 60]
        if {$ageOfDir >= $releaseMoveTime} {
            incr allDirsPass
        }
    }
    if {$allDirsPass == [llength $dirsToCheck]} {
        return 1
    }
    return 0
}

# If any wildcarded-expression in skipList matches
# the incoming dir, then return 1, else return 0.
proc isInSkipList {dir} {
    global skipList
    foreach skipMe $skipList {
        if {[string match $skipMe $dir]} {
            return 1
        }
    }
    return 0
}

# If the incoming dir has this file in it, then return 1.
# Else return 0.
proc hasSkipTag {dir} {
    global skipTag
    if {[file exists "$dir/$skipTag"]} {
        return 1
    }
    return 0
}

# Return the size of the dir tree in KB using `du`.
proc getDirSize {dir} {
    return [lindex [exec -- du -s -k "$dir"] 0]
}

# Takes in a dir and the amount of data (in KB) desired
# to be free on the dir's partition.  If available space
# exists, returns 1, else (not enough space) returns 0.
proc hasEnoughSpace {dir amountNeeded} { global actualDevices mountPoints
    # Keep track of indexes.
    set pos -1
    set mountedDevices [lrange [split [exec -- df -k] "\n"] 1 end]
    for {set a 0} {$a < [llength $mountedDevices]} {incr a} {
        # Grab the mount point from the `df -k` cmd (5th column).
        # Hopefully you don't have spaces in your mount point,
        # but just in case use lrange instead of lindex.
        if {[string equal [lrange [lindex $mountedDevices $a] 5 end] $dir]
            || [string match "[lrange [lindex $mountedDevices $a] 5 end]/*" $dir]} {
           # Store the best matched mount point.  We're guranteed
           # that the best mount point (most specific) will be down
           # the list since, for example, you can't mount /site/movies/vcd
           # before you mount /site/movies.  Thus, go through entire list
           # keeping track of the best matched mount point until the list is
           # finished.
           set pos $a
        }
    }

    set currentFree [lindex [lindex $mountedDevices $pos] 3]
    if {$currentFree < $amountNeeded} {
        # Not enough free space.
        return 0
    }
    return 1
}




# Begin main()...

# If system load average is too high, then bail.
set currentSystemLoad [lrange [exec -- uptime] end-2 end]
if {([lindex $maxSystemLoad 0] < [string range $currentSystemLoad 0 3])
    || ([lindex $maxSystemLoad 1] < [string range $currentSystemLoad 6 9])
    || ([lindex $maxSystemLoad 2] < [string range $currentSystemLoad 12 15])} {
    puts "failure: system load is too high ($currentSystemLoad)."
    exit $exit(HIGH_LOAD)
}

# Loop through all elements in moveTable.
foreach dir $moveTable {
    set dir [split $dir ":"]
    set srcDir [lindex $dir 0]
    set destDir [lindex $dir 1]
    if {!([file isdirectory $srcDir] && [file isdirectory $destDir])} {
        puts "failure: either source (\"$srcDir\") or destination (\"$destDir\") is not a dir."
        exit $exit(BAD_MOVETABLE)
    }

    puts "START: searching all dirs in \"$srcDir/*\"..."

    # Dirs that qualify to be moved.
    set candidateDirs ""
    foreach releaseDir [glob -nocomplain "$srcDir/*"] {
        if {[isOldEnough $releaseDir] && ![isInSkipList $releaseDir] && ![hasSkipTag $releaseDir]} {
            # Store as two-deminsional array: <dir> <modificationTime>
            lappend candidateDirs "\{$releaseDir\} [file mtime $releaseDir]"
        }
    }

    # Sort the qualified dirs from oldest -> newest (most recent) modification time.
    set candidateDirs [lsort -increasing -integer -index 1 $candidateDirs]

    # Get dir size in KB.
    # Loop through all candidates.
    foreach candidate $candidateDirs {
        # candidate is "<dir> <mtime>"
        # candidate is now our release dir (we don't need the mtime here).
        set candidate [lindex $candidate 0]
        # Verify the partition the destination dir resides has enough space
        # to store the candidate.
        set srcDirSize [getDirSize $candidate]
        if {![hasEnoughSpace $destDir $srcDirSize]} {
            # Since there isn't enough free space, run the free space script
            # and test again whether or not enough space exists.
            exec -- $freeSpaceScript
            if {![hasEnoughSpace $destDir $srcDirSize]} {
                puts "warning: can't move source (\"$candidate\") because destination (\"$destDir\" doesn't have enough free space."
                continue
            }
        }

        # Check if destination already exists.
        if {[file exists "$destDir/[file tail $candidate]"]} {
            if {$removeDupeDest} {
                puts "warning: found duplicate destination (\"$destDir/[file tail $candidate]\"), going to delete it and copy over the source i have"
                file delete -force -- "$destDir/[file tail $candidate]"
            } else {
                puts "warning: found duplicate destination (\"$destDir/[file tail $candidate]\"), going to skip this source (\"$candidate\")."
                continue
            }
        }

        # Lock down source dir while we copy, then bcwipe it.
	# ???

        # Copy to destination.
        puts "  -> copy [file tail $candidate] @ [expr $srcDirSize / 1024]MB"
        puts "     going from source (\"$candidate\") to destination dir (\"$destDir/\")..."
        #file copy -- $candidate $destDir
        exec -- cp -a $candidate $destDir

        # Now wipe the original with bcwipe (www.jetico.com).
        # We're going to use "U.S. DoD 5200.28 seven pass extended character rotation wiping,"
        # instead of the default 35-pass because of the large amount of data.
        puts "  x: wiping [file tail $candidate] @ [expr $srcDirSize / 1024]MB"
        puts "     source being wiped.  this roughly takes 1min per 100MB, so your estimate is [expr ($srcDirSize / 1024) / 100.0] minutes..."
        set startTime [clock seconds]
        exec -- bcwipe -Iqr $candidate
        puts "     done.  actual elapsed time: [expr ([clock seconds] - $startTime) / 60.0] minutes."

        puts "  v: release \"[file tail $candidate]\" successfully processed."
        puts "Current load average: [lrange [exec -- uptime] end-2 end]"
    }
    puts "STOP: done searching all dirs in \"$srcDir/*\".\n"
}

puts "Done with all of the move table, see ya around topgun! ;)"
exit $exit(SUCCESS)

# EOF. :-)
